#include <windows.h>
#include <d3dx8.h>
#include <dinput.h>

#include "setup.h"
#include "meshutil.h"
#include "objekty.h"

#define CLASSNAME TEXT("DirectX:Direct3D")
#define APPNAME TEXT("Test04")

struct CREATEDEVICEPARAMETERS {
	UINT Adapter;
	D3DDEVTYPE DeviceType;
	DWORD BahaviorFlags;
	D3DPRESENT_PARAMETERS PParameters;
};

void CleanManaged();
void CleanDefault();
void LoadManaged();
void LoadDefault();

void RestoreD3D();
void SwapD3DDevice();
void ResizeWindow(DWORD, DWORD);
BOOL GetD3DDeviceParameters();

HINSTANCE GhInstance;
HWND Ghwnd;

LPDIRECT3D8				GlpD3D = NULL;
LPDIRECT3DDEVICE8		GlpD3DDevice = NULL;
CREATEDEVICEPARAMETERS	GWndCDP, GFullCDP;

LPDIRECTINPUT8			GlpDI = NULL;
LPDIRECTINPUTDEVICE8	GlpDIMouse = NULL;

BOOL Gafull;
FLOAT GBBufferAspectRatio;
FLOAT GMonitorAspectRatio;
FLOAT GPersp;

BOOL GPause = FALSE;
BOOL alt = FALSE;
FLOAT v = 0;

LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam) {
	switch (msg) {
	
	case WM_DESTROY:
		PostQuitMessage(0);
		break;
	
	case WM_SYSCOMMAND:
		switch (wParam) {
		case SC_SCREENSAVE:
		case SC_MONITORPOWER:
		case SC_KEYMENU:
			return 0;
		}
		break;

	case WM_KEYDOWN:
//		if (wParam == VK_ESCAPE) PostQuitMessage(0);
		if (wParam == VK_ESCAPE) SendMessage(Ghwnd, WM_CLOSE, 0, 0);
//		if (wParam == VK_MENU) alt = TRUE;
		if (wParam == VK_RETURN) SwapD3DDevice();
//		if (wParam == VK_SPACE) GPause = !GPause;
		if (wParam == VK_SPACE) v = 0.0f;
		break;

	case WM_KEYUP:
		if (wParam == VK_F1) MessageBox(Ghwnd, TEXT("Direct 3D - Test02\nAdam(c) 2002"), APPNAME, MB_ICONINFORMATION);
//		if (wParam == VK_F1) GetD3DDeviceParameters();
//		if (wParam == VK_MENU) alt = FALSE;
		break;

	case WM_SIZE:
			ResizeWindow(LOWORD(lParam),HIWORD(lParam));
			return 0;

	}

	return DefWindowProc(hwnd, msg, wParam, lParam);

}


HRESULT CreateD3DDevice(CREATEDEVICEPARAMETERS *cdp) {
	return GlpD3D->CreateDevice(
		cdp->Adapter,
		cdp->DeviceType,
		Ghwnd,
		cdp->BahaviorFlags,
		&cdp->PParameters,
		&GlpD3DDevice);
}

VOID ResizeWindow(DWORD Width, DWORD Height) {
	if (!Gafull && GlpD3DDevice != NULL) {
		if (Height <= 0L) Height = 1L;
		if (Width <= 0L) Width = 1L;
		GWndCDP.PParameters.BackBufferWidth = Width;
		GWndCDP.PParameters.BackBufferHeight = Height;
		GlpD3DDevice->Reset(&GWndCDP.PParameters);
	}
}

VOID SwapD3DDevice() {
	Gafull = !Gafull;
	RestoreD3D();

	if (!Gafull) {
		RECT WindowRect;
		WindowRect.left = 10L;
		WindowRect.top = 20L;
		WindowRect.right = GWndCDP.PParameters.BackBufferWidth;
		WindowRect.bottom = GWndCDP.PParameters.BackBufferHeight;
		MoveWindow(Ghwnd, WindowRect.left, WindowRect.top, WindowRect.right, WindowRect.bottom, TRUE);
	}
}

HRESULT InitD3D() {
	if (NULL == (GlpD3D = Direct3DCreate8(D3D_SDK_VERSION)) )
		return E_FAIL;

	if (Gafull)
		if (FAILED(CreateD3DDevice(&GFullCDP))) return E_FAIL;
	if (!Gafull)
		if (FAILED(CreateD3DDevice(&GWndCDP))) return E_FAIL;


	return S_OK;
}

HRESULT InitDInput() {
	if (FAILED(DirectInput8Create(GetModuleHandle(NULL), DIRECTINPUT_VERSION, IID_IDirectInput8, (VOID **)&GlpDI, NULL)))
		return E_FAIL;

	if (FAILED(GlpDI->CreateDevice(GUID_SysMouse, &GlpDIMouse, NULL))) 
		return E_FAIL;

	GlpDIMouse->SetDataFormat(&c_dfDIMouse);
//	GlpDIMouse->SetCooperativeLevel(Ghwnd, DISCL_NONEXCLUSIVE | DISCL_BACKGROUND);
//	GlpDIMouse->SetCooperativeLevel(Ghwnd, DISCL_NONEXCLUSIVE | DISCL_FOREGROUND);
	GlpDIMouse->SetCooperativeLevel(Ghwnd, DISCL_EXCLUSIVE | DISCL_FOREGROUND);

	if (FAILED(GlpDIMouse->Acquire()))
		return E_FAIL;

	return S_OK;
}

VOID CleanDInput() {
	if (GlpDIMouse != NULL) {
		GlpDIMouse->Unacquire();
		GlpDIMouse->Release();
	}
	if (GlpDI != NULL) GlpDI->Release();
}

VOID CleanD3D() {
	CleanManaged();
	CleanDefault();
	if (GlpD3DDevice != NULL) 
		GlpD3DDevice->Release();
	if (GlpD3D != NULL) {
		GlpD3D->Release();
	}
}

VOID RestoreD3D() {
	CleanDefault();

	if (Gafull) GlpD3DDevice->Reset(&GFullCDP.PParameters);
	if (!Gafull) GlpD3DDevice->Reset(&GWndCDP.PParameters);

	if (Gafull) GBBufferAspectRatio = GMonitorAspectRatio;
	if (!Gafull) GBBufferAspectRatio = (FLOAT)GWndCDP.PParameters.BackBufferWidth/(FLOAT)GWndCDP.PParameters.BackBufferHeight;

	LoadDefault();
}


Zeme *Zeme0;
Kamera *Kamera0;

VOID InitLights() {
	D3DXVECTOR3 vecDir;
	D3DLIGHT8 light;
	ZeroMemory(&light, sizeof(D3DLIGHT8));

	light.Type = D3DLIGHT_POINT;
	light.Diffuse.r = 1.0f;
	light.Diffuse.g = 1.0f;
	light.Diffuse.b = 0.9f;
	light.Specular.r = 1.0f;
	light.Specular.g = 1.0f;
	light.Specular.b = 1.0f;
	light.Ambient.r = 0.0f;
	light.Ambient.g = 0.0f;
	light.Ambient.b = 0.0f;
	light.Position = D3DXVECTOR3(-26.0f, -24.0f, -3.0f);

	light.Range = 1000.0f;
	light.Attenuation0 = 1.0f;
	light.Attenuation1 = 0.0f;
	light.Attenuation2 = 0.0f;

	GlpD3DDevice->SetLight(0, &light);
	GlpD3DDevice->LightEnable(0, TRUE);
	GlpD3DDevice->SetRenderState(D3DRS_LIGHTING, TRUE);
	GlpD3DDevice->SetRenderState(D3DRS_AMBIENT, 0x00808080);
}

VOID InitMatrices() {
}

VOID InitGeometry() {
	Zeme0 = new Zeme(GlpD3DDevice);
	Kamera0 = new Kamera(GlpD3DDevice, &GPersp, &GBBufferAspectRatio);
	Kamera0->Pozice.x = 0.0f;
	Kamera0->Pozice.y = 0.0f;
	Kamera0->Pozice.z = -4.0f;
	Kamera0->Rychlost.x = 0.0f;
	Kamera0->Rychlost.y = 0.0f;
	Kamera0->Rychlost.z = 0.0f;
	D3DXMatrixIdentity(&(Kamera0->matCamRot));
}

VOID LoadDefault() {
	Zeme0->LoadDefault();
	Kamera0->LoadDefault();

	InitLights();
}

VOID LoadManaged() {
	Zeme0->LoadManaged();

}

VOID CleanDefault() {
	Zeme0->CleanDefault();
}

VOID CleanManaged() {
	Zeme0->CleanManaged();
}

DIMOUSESTATE MouseState;
VOID Drive() {

	FLOAT alpha, gamma, beta;
    GlpDIMouse->GetDeviceState(sizeof(MouseState),(LPVOID)&MouseState);
    gamma = (FLOAT)MouseState.lY*0.005f;
    alpha = (FLOAT)MouseState.lX*0.005f;
    beta = (FLOAT)MouseState.lZ*0.0005f;

	D3DXMATRIX matW1, matW2;
	D3DXVECTOR3 l, b, t;

	l.x = Kamera0->matCamRot._11;
	l.y = Kamera0->matCamRot._12;
	l.z = Kamera0->matCamRot._13;

	b.x = Kamera0->matCamRot._21;
	b.y = Kamera0->matCamRot._22;
	b.z = Kamera0->matCamRot._23;

	t.x = Kamera0->matCamRot._31;
	t.y = Kamera0->matCamRot._32;
	t.z = Kamera0->matCamRot._33;


	D3DXMatrixRotationAxis(&matW1, &t, beta);
	D3DXMatrixMultiply(&(Kamera0->matCamRot), &(Kamera0->matCamRot), &matW1);
	D3DXMatrixRotationAxis(&matW1, &b, alpha);
	D3DXMatrixMultiply(&(Kamera0->matCamRot), &(Kamera0->matCamRot), &matW1);
	D3DXMatrixRotationAxis(&matW1, &l, gamma);
	D3DXMatrixMultiply(&(Kamera0->matCamRot), &(Kamera0->matCamRot), &matW1);

	if (MouseState.rgbButtons[1])
//		Kamera0->Rychlost.x -= t.x * 0.001f;
//		Kamera0->Rychlost.y -= t.y * 0.001f;
//		Kamera0->Rychlost.z -= t.z * 0.001f;
		v -= 0.0001f;
//		v -= 1.0f;

	if (MouseState.rgbButtons[0])
//		Kamera0->Rychlost.x += t.x * 0.001f;
//		Kamera0->Rychlost.y += t.y * 0.001f;
//		Kamera0->Rychlost.z += t.z * 0.001f;
		v += 0.0001f;
//		v += 1.0f;

	Kamera0->Rychlost = t * v;

	Kamera0->Pozice += Kamera0->Rychlost;
	D3DXMatrixTranslation(&(Kamera0->matCamTrn), Kamera0->Pozice.x, Kamera0->Pozice.y, Kamera0->Pozice.z);
}

VOID Render() {
//	GlpD3DDevice->Clear(0, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(255, 255, 255), 1.0f, 0);
	GlpD3DDevice->Clear(0, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0, 0, 0), 1.0f, 0);
	GlpD3DDevice->BeginScene();

	Kamera0->Zapni();

	Zeme0->Render(&(Kamera0->matCamTrn));

	
	GlpD3DDevice->EndScene();
	GlpD3DDevice->Present(NULL, NULL, NULL, NULL);
}

BOOL CreateWin() {
	WNDCLASSEX wc;
	wc.cbSize = sizeof(WNDCLASSEX);
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
	wc.hIconSm = LoadIcon(NULL, IDI_APPLICATION);
	wc.hInstance = GhInstance;
	wc.lpfnWndProc = WndProc;
	wc.lpszClassName = CLASSNAME;
	wc.lpszMenuName = NULL;
	wc.style = CS_HREDRAW | CS_VREDRAW;

	if (!RegisterClassEx(&wc)) return FALSE;

	RECT WindowRect;
	WindowRect.left = 10L;
	WindowRect.top = 20L;
	WindowRect.right = GWndCDP.PParameters.BackBufferWidth;
	WindowRect.bottom = GWndCDP.PParameters.BackBufferHeight;

	Ghwnd = CreateWindowEx(0,
		CLASSNAME, APPNAME,
		WS_OVERLAPPEDWINDOW | WS_VISIBLE,
		WindowRect.left, WindowRect.top, WindowRect.right, WindowRect.bottom,
		NULL, NULL, GhInstance, NULL);
	
	if (!Ghwnd) return FALSE;

	return TRUE;
}

BOOL GetD3DDeviceParameters() {
	SETUPINFO SetupInfo;
	if (!Setup(GhInstance, &SetupInfo)) return FALSE;

	Gafull = SetupInfo.FullScreen;

	ZeroMemory(&GWndCDP.PParameters, sizeof(GWndCDP.PParameters));
	ZeroMemory(&GFullCDP.PParameters, sizeof(GFullCDP.PParameters));


	GWndCDP.PParameters.Windowed = TRUE;
	GWndCDP.PParameters.BackBufferWidth = 400;
	GWndCDP.PParameters.BackBufferHeight = 400;
	GWndCDP.PParameters.BackBufferFormat = SetupInfo.ActualDisplay.Format;
	GWndCDP.PParameters.SwapEffect = D3DSWAPEFFECT_DISCARD;
	GWndCDP.PParameters.EnableAutoDepthStencil = TRUE;
	GWndCDP.PParameters.AutoDepthStencilFormat = D3DFMT_D16;
	GWndCDP.Adapter = D3DADAPTER_DEFAULT;
	GWndCDP.DeviceType = D3DDEVTYPE_HAL;
	GWndCDP.BahaviorFlags = D3DCREATE_SOFTWARE_VERTEXPROCESSING;

	
	GFullCDP.PParameters.Windowed = FALSE;
	GFullCDP.PParameters.BackBufferWidth = SetupInfo.Display.Width;
	GFullCDP.PParameters.BackBufferHeight = SetupInfo.Display.Height;
	GFullCDP.PParameters.BackBufferFormat = SetupInfo.Display.Format;
	GFullCDP.PParameters.SwapEffect = D3DSWAPEFFECT_DISCARD;
	GFullCDP.PParameters.EnableAutoDepthStencil = TRUE;
	GFullCDP.PParameters.AutoDepthStencilFormat = D3DFMT_D16;
	GFullCDP.Adapter = SetupInfo.Adapter;
	GFullCDP.DeviceType = D3DDEVTYPE_HAL;
	GFullCDP.BahaviorFlags = D3DCREATE_SOFTWARE_VERTEXPROCESSING;


//	GD3Dpp.BackBufferFormat = D3DFMT_A8R8G8B8 ;
//	GD3Dpp.AutoDepthStencilFormat = D3DFMT_D32;

//	GD3Dpp.MultiSampleType = D3DMULTISAMPLE_2_SAMPLES;

	GMonitorAspectRatio = 4.0f/3.0f;
	GPersp = D3DX_PI*0.25f;

	if (Gafull) GBBufferAspectRatio = GMonitorAspectRatio;
	if (!Gafull) GBBufferAspectRatio = (FLOAT)GWndCDP.PParameters.BackBufferWidth/(FLOAT)GWndCDP.PParameters.BackBufferHeight;

	return TRUE;
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd) {
	
	GhInstance = hInstance;

	if (!GetD3DDeviceParameters()) return 0;

	if (!CreateWin()) return 0;

	if (!SUCCEEDED(InitD3D())) {
		MessageBox(NULL, TEXT("Nepodailo se inicializovat Direct3D"), APPNAME, MB_ICONERROR);
		return 0;
	}

	InitGeometry();
	InitLights();
	InitMatrices();

	if (!SUCCEEDED(InitDInput())) {
		MessageBox(NULL, TEXT("Nepodailo se inicializovat DirectInput"), APPNAME, MB_ICONERROR);
		return 0;
	}

	MSG msg;
	BOOL gm = TRUE;
	while (gm) {
		if(PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE)) {
			if(gm = GetMessage(&msg, NULL, 0, 0)) {
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}
		else {
			if (GlpD3DDevice->TestCooperativeLevel() == D3D_OK) {
				if (!GPause) {
					Drive();
					Render();
				}
			} else {
				if (GlpD3DDevice->TestCooperativeLevel() == D3DERR_DEVICENOTRESET) RestoreD3D();
			}
		}
	}

	CleanDInput();
	CleanD3D();

	return msg.wParam;
}